import QtQuick 2.3
import Fly.Components 1.0

Item {
    id: selectItem

    readonly property real  itemWidth:  root.itemHeight * 0.7
    readonly property real  xOffset:    flickable && selectionVisible ? itemWidth : 0
    readonly property real  yOffset:    flickable ? flickable.contentY : 0
    readonly property bool  isDragging: (selectionLoader.item && selectionLoader.item.dragging)
                                        || (cursorLoader.item && cursorLoader.item.dragging)

    property bool textChangedByOutside: true
    property bool selectionVisible:     true

    property var  flickable:            null
    property var  inputTarget:          null

    visible: !!inputTarget
    enabled: !isDesktopMode
    z: 9999

    // TODO Text
    // TODO vibro on show
    // TODO when dragging selection in flickable, don't show selection in magnifier on boundaries
    // TODO test as separate widgets, e.g. without FlyRoot

    function show(inputTarget, flickable, selectionVisible) {
        if(!inputTarget) return;
        if(selectItem.inputTarget === inputTarget && selectItem.selectionVisible === selectionVisible) {
            if(cursorLoader.item) cursorLoader.item.hideCursorTimer.restart()
            return;
        }

        selectItem.inputTarget = inputTarget;
        selectItem.flickable = flickable;
        selectItem.selectionVisible = selectionVisible;
        selectItem.width = flickable ? flickable.width + (selectionVisible ? itemWidth * 2 : 0) : inputTarget.width;
        selectItem.height = flickable ? flickable.height + itemWidth : inputTarget.height;

        var newPos = mapFromItem(flickable ? flickable : inputTarget, x, y);
        selectItem.x = newPos.x - xOffset;
        selectItem.y = newPos.y;

        if(selectionVisible) selectionDbusNotifier.emitVisibleChanged(true)
    }

    function done() {
        inputTarget.deselect();
        inputTarget = null
        flickable = null

        if(selectionVisible) selectionDbusNotifier.emitVisibleChanged(false)
    }

    Connections {
        target: inputTarget
        onActiveFocusChanged: if(inputTarget && !inputTarget.activeFocus) done()
        onLengthChanged: if(inputTarget && inputTarget.activeFocus && textChangedByOutside && !isDragging) done()
    }

    Connections {
        target: flickable
        onMovementStarted: if(inputTarget && !selectionVisible && !isDragging) done()
    }

    FlySelectionDbusNotifier {
        id: selectionDbusNotifier
        onCut: if(inputTarget) { textChangedByOutside = false; inputTarget.cut(); done(); textChangedByOutside = true }
        onCopy: if(inputTarget) { inputTarget.copy(); done() }
        onPaste: if(inputTarget) { textChangedByOutside = false; inputTarget.paste(); done(); textChangedByOutside = true }
        onSelectAll: if(inputTarget) { inputTarget.selectAll() }
        onVisibleChanged: if(!visible && inputTarget && inputTarget.activeFocus) done()
    }

    Loader {
        id: selectionLoader

        readonly property rect  startSelCoords: selectItem.inputTarget
                                                ? selectItem.inputTarget.positionToRectangle(selectItem.inputTarget.selectionStart)
                                                : Qt.rect(0,0,0,0)
        readonly property rect  endSelCoords:   selectItem.inputTarget
                                                ? selectItem.inputTarget.positionToRectangle(selectItem.inputTarget.selectionEnd)
                                                : Qt.rect(0,0,0,0)

        width: parent.width; height: parent.height
        active: inputTarget && selectionVisible && (inputTarget.selectionStart !== inputTarget.selectionEnd)
        enabled: active
        clip: !!flickable

        function updateSelectionStart(x, y) {
            var pos = Math.min(selectItem.inputTarget.selectionEnd - 1,
                               selectItem.inputTarget.positionAt(x - selectItem.xOffset, y + selectItem.yOffset));
            if(pos !== selectItem.inputTarget.selectionStart && pos < selectItem.inputTarget.selectionEnd)
                selectItem.inputTarget.select(pos, selectItem.inputTarget.selectionEnd)
        }

        function updateSelectionEnd(x, y) {
            var pos = Math.max(selectItem.inputTarget.selectionStart + 1,
                               selectItem.inputTarget.positionAt(x - selectItem.xOffset, y + selectItem.yOffset));
            if(pos !== selectItem.inputTarget.selectionEnd && pos > selectItem.inputTarget.selectionStart)
                selectItem.inputTarget.select(selectItem.inputTarget.selectionStart, pos)
        }

        sourceComponent: Item {
            id: selectionItem

            readonly property bool dragging:        startSelMouseArea.drag.active || endSelMouseArea.drag.active
            readonly property bool draggingStart:   startSelMouseArea.drag.active

            property Translate translateStart: Translate {
                x: -startSelImage.width + selectItem.itemWidth * 0.16;
                y: selectionLoader.startSelCoords.height / 2
            }
            property Translate translateEnd:   Translate {
                x: selectionLoader.endSelCoords.width - selectItem.itemWidth * 0.16;
                y: selectionLoader.startSelCoords.height / 2
            }

            FlyColoredImage {
                id: startSelImage
                x: selectionLoader.startSelCoords.x + selectItem.xOffset;
                y: selectionLoader.startSelCoords.y - selectItem.yOffset + selectionLoader.startSelCoords.height / 2
                width: selectItem.itemWidth; height: width
                transform: selectionItem.translateStart
                imageSourceSize: Qt.size(selectItem.itemWidth + 10, selectItem.itemWidth + 10)
                imageShadow: true
                imagePath: "qrc:/images/start-selection.svg"
                imageColorEnabled: true
                imageColor: styles.highlightColor
            }

            FlyColoredImage {
                id: endSelImage
                x: selectionLoader.endSelCoords.x + selectItem.xOffset;
                y: selectionLoader.endSelCoords.y - selectItem.yOffset + selectionLoader.endSelCoords.height / 2
                width: selectItem.itemWidth; height: width
                transform: selectionItem.translateEnd
                imageSourceSize: Qt.size(selectItem.itemWidth + 10, selectItem.itemWidth + 10)
                imageShadow: true
                imagePath: "qrc:/images/end-selection.svg"
                imageColorEnabled: true
                imageColor: styles.highlightColor
            }

            MouseArea {
                id: startSelMouseArea
                width: startSelImage.width; height: startSelImage.height
                transform: selectionItem.translateStart
                drag {
                    target: startSelMouseArea
                    axis: Drag.XAndYAxis
                    minimumX: 0; maximumX: selectItem.width - startSelImage.width
                    minimumY: 0; maximumY: selectItem.height - startSelImage.height
                }
                onXChanged: if(drag.active) selectionLoader.updateSelectionStart(x, y)
                onYChanged: if(drag.active) selectionLoader.updateSelectionStart(x, y)

                Binding on x { when: !startSelMouseArea.drag.active; value: startSelImage.x }
                Binding on y { when: !startSelMouseArea.drag.active; value: startSelImage.y }
            }

            MouseArea {
                id: endSelMouseArea
                width: endSelImage.width; height: endSelImage.height
                transform: selectionItem.translateEnd
                drag {
                    target: endSelMouseArea
                    axis: Drag.XAndYAxis
                    minimumX: 0; maximumX: selectItem.width
                    minimumY: 0; maximumY: selectItem.height - endSelImage.height
                }
                onXChanged: if(drag.active) selectionLoader.updateSelectionEnd(x, y)
                onYChanged: if(drag.active) selectionLoader.updateSelectionEnd(x, y)

                Binding on x { when: !endSelMouseArea.drag.active; value: endSelImage.x }
                Binding on y { when: !endSelMouseArea.drag.active; value: endSelImage.y }
            }
        }
    }

    Loader {
        id: cursorLoader
        width: parent.width; height: parent.height
        active: inputTarget && !selectionVisible
        enabled: active

        function updateCursorPosition(x, y) {
            var pos = selectItem.inputTarget.positionAt(x, y + selectItem.yOffset);
            if(pos !== selectItem.inputTarget.cursorPosition) {
                inputTarget.select(pos, pos + 1); // HACK for vkbd prediction
                inputTarget.select(pos, pos);
            }
        }

        sourceComponent: Item {
            id: cursorItem

            readonly property alias hideCursorTimer: hideCursorTimer
            readonly property bool dragging:    cursorMouseArea.drag.active
            property Translate translateOffset: Translate {
                x: -cursorImage.width / 2
                y: selectItem.inputTarget ? selectItem.inputTarget.cursorRectangle.height / 2 : 0
            }

            Timer {
                id: hideCursorTimer
                interval: 3000
                running: cursorLoader.active && !cursorItem.dragging
                onTriggered: selectItem.done()
            }

            FlyColoredImage {
                id: cursorImage
                x: selectItem.inputTarget ? selectItem.inputTarget.cursorRectangle.x : 0
                y: selectItem.inputTarget
                   ? selectItem.inputTarget.cursorRectangle.y + selectItem.inputTarget.cursorRectangle.height / 2 - selectItem.yOffset
                   : 0
                width: selectItem.itemWidth; height: width
                transform: cursorItem.translateOffset
                imageSourceSize: Qt.size(selectItem.itemWidth + 10, selectItem.itemWidth + 10)
                imageShadow: true
                imagePath: "qrc:/images/cur-selection.svg"
                imageColorEnabled: true
                imageColor: styles.highlightColor
            }

            MouseArea {
                id: cursorMouseArea
                width: cursorImage.width; height: cursorImage.height
                transform: cursorItem.translateOffset
                drag {
                    target: cursorMouseArea
                    axis: Drag.XAndYAxis
                    minimumX: 0; maximumX: selectItem.width
                    minimumY: 0; maximumY: selectItem.height - cursorImage.height
                }
                onXChanged: if(drag.active) { cursorLoader.updateCursorPosition(x, y) }
                onYChanged: if(drag.active) { cursorLoader.updateCursorPosition(x, y) }

                Binding on x { when: !cursorMouseArea.drag.active; value: cursorImage.x }
                Binding on y { when: !cursorMouseArea.drag.active; value: cursorImage.y }
            }
        }
    }

    FlyShadow { target: magnifier }
    Rectangle {
        id: magnifier

        readonly property int charCount: 8
        readonly property int charCountHalf: charCount / 2
        readonly property string maskChar: ' '
        readonly property string mask: [maskChar].map(function(val) { // no repeat func :(
            var t = "";
            for(var i = 0; i < charCountHalf; i++) t += val;
            return t;
        })[0]

        x: visible ? getX() : 0
        y: visible ? getY() : 0
        width: inputTarget ? inputTarget.font.pixelSize * charCount : 1
        height: root.itemHeight
        color: styles.colors.white
        border { color: styles.colors.gray; width: 1 }
        radius: 4
        clip: true
        visible: isDragging
        enabled: visible

        function getX() {
            var x = selectionVisible
                    ? (inputTarget.positionToRectangle(selectionLoader.item.draggingStart
                                                       ? inputTarget.selectionStart
                                                       : inputTarget.selectionEnd).x) + itemWidth
                    : inputTarget.cursorRectangle.x;
            x -= width / 2;
            return Math.max(-selectItem.x, Math.min(selectItem.parent.width - selectItem.x - width, x));
        }

        function getY() {
            var y = selectionVisible
                    ? (inputTarget.positionToRectangle(selectionLoader.item.draggingStart
                                                       ? inputTarget.selectionStart
                                                       : inputTarget.selectionEnd).y)
                    : inputTarget.cursorRectangle.y;
            y -= height + 10 * dpiRatio
            return Math.max(-selectItem.y, Math.min(selectItem.parent.height - selectItem.y - height, y));
        }

        TextInput {
            id: magnifierInput
            anchors.fill: parent
            activeFocusOnPress: false
            font.pixelSize: parent.height * 0.7
            horizontalAlignment: TextInput.AlignHCenter; verticalAlignment: TextInput.AlignVCenter
            color: styles.lineedit.textColor
            selectedTextColor: styles.label.color
            selectionColor: styles.highlightColor2
            text: visible ? getText() : ""
            cursorVisible: cursorPosition < magnifier.charCount
            cursorPosition: magnifier.charCountHalf

            function getText() {
                var pos = selectionVisible
                        ? (selectionLoader.item.draggingStart ? inputTarget.selectionStart : inputTarget.selectionEnd)
                        : inputTarget.cursorPosition;
                var posY = inputTarget.positionToRectangle(pos).y;
                var left = inputTarget.getText(pos - magnifier.charCountHalf, pos);
                var right = inputTarget.getText(pos, pos + magnifier.charCountHalf);
                var i = 0;

                if(flickable) {
                    for(i = 0; i < left.length; i++) {
                        if(posY === inputTarget.positionToRectangle(pos - magnifier.charCountHalf + i).y)
                            break;
                        left = left.substring(0, i) + magnifier.maskChar + left.substring(i + 1);
                    }

                    for(i = right.length - 1; i >= 0; i--) {
                        if(posY === inputTarget.positionToRectangle(pos + i + 1).y)
                            break;
                        right = right.substring(0, i) + magnifier.maskChar + right.substring(i + 1);
                    }
                }

                if(left.length < magnifier.charCountHalf) left = (magnifier.mask + left).slice(-magnifier.charCountHalf);
                if(right.length < magnifier.charCountHalf) right = (right + magnifier.mask).slice(0, magnifier.charCountHalf);

                return left + right;
            }

            onTextChanged: {
                cursorPosition = magnifier.charCountHalf;

                if(selectionVisible && inputTarget.selectionStart !== inputTarget.selectionEnd) {
                    if(selectionLoader.item.draggingStart) {
                        magnifierInput.select(magnifier.charCountHalf,
                                              Math.min(magnifier.charCountHalf + inputTarget.selectionEnd - inputTarget.selectionStart,
                                                       magnifier.charCount))
                    } else {
                        magnifierInput.select(Math.max(magnifier.charCountHalf - (inputTarget.selectionEnd - inputTarget.selectionStart), 0),
                                              magnifier.charCountHalf)
                    }
                }
            }
        }
    }
}
